/*
    SmartHomeGeeks.io Smart Garage Door
    https://smarthomegeeks.io/smart-garage-door/
*/

#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <ArduinoOTA.h>
#include <ArduinoJson.h>
#include <arduino_secrets.h>

// Device configuration
const unsigned int relay = 4;                               // relay port
const unsigned int sensor = 5;                              // stop IO port
const unsigned int moveTime = 25000;                        // time in milliseconds it takes to open or close garage door
const unsigned int pushButtonTime = 200;                    // time in milliseconds to press a button
const String deviceName = "Garage Door";                     // device name

// WiFi configuration
const char* ssid = SECRET_SSID;
const char* password = SECRET_PASS;

// MQTT configuration
const String mqttClient = "garage_door";                    // MQTT client name
const String mqttStateTopic = mqttClient + "/state";        // MQTT state topic
const String mqttCmdTopic = mqttClient + "/command";        // MQTT command topic
const String mqttPosTopic = mqttClient + "/position";       // MQTT position topic
const String mqttAvlTopic = mqttClient + "/availability";   // MQTT position topic
const char* mqttServer = SECRET_MQTT_BROKER;                // MQTT broker IP
const char* mqttUser = SECRET_MQTT_USERNAME;                // MQTT username
const char* mqttPassword = SECRET_MQTT_PASSWORD;            // MQTT password
const int mqttPort = 1883;                                  // MQTT port (default: 1883)

// Create some global variables
unsigned long lastMQTTconnect = 0;
unsigned long timer = 0;
unsigned int prevSensor;
char charPayload[5];
String lastCommand;

WiFiClient espClient;
PubSubClient client(espClient);

void setup_wifi() {
  delay(1000);

  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.print(ssid);

  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  Serial.println();
  Serial.println("WiFi connected");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
}

void setup() {
  pinMode(relay, OUTPUT);

  Serial.begin(115200);

  setup_wifi();

  // Configure OTA
  ArduinoOTA.setHostname(mqttClient.c_str());
  ArduinoOTA.setPassword(SECRET_OTA_PASSWORD);
  ArduinoOTA.begin();

  // Configure MQTT
  client.setServer(mqttServer, mqttPort);
  client.setCallback(callback);
  connectmqtt();

  Serial.print("Last Command: ");
  Serial.println(lastCommand);

  client.publish(mqttStateTopic.c_str(), (digitalRead(sensor) ? "closed" : "open"), true);
  client.publish(mqttPosTopic.c_str(), (digitalRead(sensor) ? "0" : "50"), true);
  lastCommand = (digitalRead(sensor) ? "closed" : "open");
}

void connectmqtt() {
  // prevent constant reconnect
  if (millis() - lastMQTTconnect < 5000) return;
  lastMQTTconnect = millis();

  if (client.connect(mqttClient.c_str(), mqttUser, mqttPassword, mqttAvlTopic.c_str(), 1, true, "offline")) {
    Serial.println("connected to MQTT");
    client.setBufferSize(512);

    // Once connected, enable auto discovery for HA
    String discoveryTopic = "homeassistant/cover/" + mqttClient + "/config";

    char buffer[512];

    DynamicJsonDocument deviceData(512);
    deviceData["name"] = deviceName;
    deviceData["mdl"] = "Smart Garage Door";
    deviceData["ids"] = mqttClient;
    deviceData["mf"] = "SmartHomeGeeks.io";
    deviceData["sw"] = "2023.08";
  
    DynamicJsonDocument root(512);
    root["name"] = nullptr;
    root["obj_id"] = mqttClient;
    root["dev_cla"] = "garage";
    root["stat_t"] = mqttStateTopic;
    root["cmd_t"] = mqttCmdTopic;
    root["avty_t"] = mqttAvlTopic;
    root["uniq_id"] = "cover." + mqttClient;
    root["pos_t"] = mqttPosTopic;
    for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
    serializeJson(root, buffer);
    client.publish(discoveryTopic.c_str(), buffer, true);
    root.clear();

    // RSSI sensor
    root["name"] = "RSSI";
    root["obj_id"] = mqttClient + "_rssi";
    root["stat_t"] = mqttClient + "/rssi";
    root["unit_of_meas"] = "dBm";
    root["ic"] = "mdi:access-point";
    root["uniq_id"] = "sensor." + mqttClient + "_rssi";
    root["avty_t"] = mqttAvlTopic;
    root["ent_cat"] = "diagnostic";
    for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
    serializeJson(root, buffer);
    client.publish(("homeassistant/sensor/" + mqttClient + "/rssi/config").c_str(), buffer, true);
    root.clear();

    // IP
    root["name"] = "IP";
    root["obj_id"] = mqttClient + "_ip";
    root["stat_t"] = mqttClient + "/ip";
    root["ic"] = "mdi:ip-network";
    root["uniq_id"] = "sensor." + mqttClient + "_ip";
    root["avty_t"] = mqttAvlTopic;
    root["ent_cat"] = "diagnostic";
    for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
    serializeJson(root, buffer);
    client.publish(("homeassistant/sensor/" + mqttClient + "/ip/config").c_str(), buffer, true);
    root.clear();

    // ... and resubscribe
    client.subscribe(mqttCmdTopic.c_str());
    client.publish(mqttAvlTopic.c_str(), "online", true);
    client.publish((String(mqttClient) + "/ip").c_str(),  WiFi.localIP().toString().c_str(), true);
    client.publish((String(mqttClient) + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
  } else {
    Serial.print("failed, rc=");
    Serial.print(client.state());
    Serial.println(" try again in 5 seconds");
  }
}

void toggle() {
    digitalWrite(relay, HIGH);
    delay(pushButtonTime);
    digitalWrite(relay, LOW);
    delay(1000);
}

void callback(char* topic, byte* payload, unsigned int length) {
  Serial.print("Message arrived: ");
  for (int i=0;i<length;i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();

  if (!strncmp((char*)payload, "OPEN", length)) {
    if (timer != 0 && lastCommand == "open") return;
    Serial.println("Opening Garage Door");

    if (timer == 0 && lastCommand == "open") {
      toggle();
      toggle();
    } else if (timer != 0 && lastCommand == "close") {
      toggle();
      client.publish(mqttStateTopic.c_str(), "stopped", true);
    }
    client.publish(mqttStateTopic.c_str(), "opening", true);
    toggle();
    timer = millis();
    lastCommand = "open";
  } else if (!strncmp((char*)payload, "CLOSE", length)) {
    if (timer != 0 && lastCommand == "close") return;
    Serial.println("Closing Garage Door");

    if (timer == 0 && lastCommand == "close") {
      toggle();
      toggle();
    } else if (timer != 0 && lastCommand == "open") {
      toggle();
      client.publish(mqttStateTopic.c_str(), "stopped", true);
    }
    client.publish(mqttStateTopic.c_str(), "closing", true);
    toggle();
    timer = millis();
    lastCommand = "close";
  } else if (!strncmp((char*)payload, "STOP", length) && timer != 0) {
    timer = 0;
    Serial.println("Stopping Garage Door");
    toggle();
    client.publish(mqttPosTopic.c_str(), "50", true);
    client.publish(mqttStateTopic.c_str(), "stopped", true);
  } 
}

void loop() {
  ArduinoOTA.handle();

  if (!client.connected()) connectmqtt();

  client.loop();

  if (digitalRead(sensor) != prevSensor) {
    prevSensor = digitalRead(sensor);
    if (prevSensor == 1) {
      client.publish(mqttStateTopic.c_str(), "closed", true);
      client.publish(mqttPosTopic.c_str(), "0", true);
      lastCommand = "close";
      timer = 0;
    } else {
      client.publish(mqttStateTopic.c_str(), "opening", true);
      client.publish(mqttPosTopic.c_str(), "50", true);
      timer = millis();
      lastCommand = "open";
    }

    client.publish((String(mqttClient) + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
  }

  if (timer != 0 && (millis() - timer) > moveTime) {
    timer = 0;
    if (lastCommand == "close") {
      toggle();
      timer = millis();
      client.publish(mqttStateTopic.c_str(), "closing", true);
      client.publish(mqttPosTopic.c_str(), "50", true);
    } else {
      client.publish(mqttStateTopic.c_str(), "open", true);
      client.publish(mqttPosTopic.c_str(), "100", true);
    }
  }
}  
